/*
 * PLUG Event Daemon
 * Fabien Devaux - 2005
 * fab AT gnux . info
 */

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <poll.h>
#include <sys/queue.h>
#include <ctype.h>
#include <stdint.h>
#include "pluged.h"

/*
 * Structures
 */

struct device_infos {
    usb_event_cookie_t cookie;
    char product[USB_MAX_STRING_LEN];
    char devname[USB_MAX_DEVNAMELEN];
	uint16_t vendor_no;
	uint16_t product_no;
	uint16_t release_no;
    SLIST_ENTRY(device_infos) next;
};

enum action_type {
	ACTION_NONE,
	ACTION_ATTACH,
	ACTION_DETACH
};

/*
 * Globals
 */

#define USB_DEVICE "/dev/usb"
int debug=0;

/*
 * Functions
 */

void match_device( struct device_infos *dev_infos , struct usb_config_head *conf, enum action_type type )
{
	struct usb_config_entry *e;
	dprint("Finding matching entries for \"%s\"\n",dev_infos->product);
	dprint("\t\tdev (%s)\n",dev_infos->devname);
	dprint("\t\tvendorID (%X)\n",dev_infos->vendor_no);
	dprint("\t\tproductID (%X)\n",dev_infos->product_no);
	dprint("\t\trelease (%X)\n",dev_infos->release_no);

	SLIST_FOREACH( e, conf, next) {
		int matched=1;

		if(e->used_fields & USEFIELD_DEV && 0 != regexec(&(e->dev_regex),dev_infos->devname,0,NULL,0))
			matched=0;
		if(e->used_fields & USEFIELD_VENDOR_NO && e->vendor_no != dev_infos->vendor_no)
			matched=0;
		if(e->used_fields & USEFIELD_PRODUCT_NO && e->product_no != dev_infos->product_no)
			matched=0;
		if(e->used_fields & USEFIELD_RELEASE_NO && e->release_no != dev_infos->release_no)
			matched=0;
		if(e->used_fields & USEFIELD_PRODUCT && 0 != regexec(&(e->product_regex),dev_infos->product,0,NULL,0))
			matched=0;

		if(matched) {
			unsigned int numstart;
			for(numstart=0; !isdigit(dev_infos->devname[numstart]) && numstart < strlen(dev_infos->devname) ;++numstart);
			dprint("%s matched!\n",e->descr);
			if( -1 == setenv("DEVNAME",dev_infos->devname,1))
				perror("setenv(DEVNAME)");
			if( -1 == setenv("DEVNUM",dev_infos->devname+numstart,1))
				perror("setenv(DEVNUM)");
			switch(type) {
				case ACTION_ATTACH:
					dprint("Attach command: %s\n",e->attach);
					system(e->attach);
					break;
				case ACTION_DETACH:
					dprint("Detach command: %s\n",e->detach);
					system(e->detach);
					break;
				case ACTION_NONE:
					break;
			}
		}
	}
}

void event_check_loop(int fd,struct usb_config_head *conf)
{
    struct pollfd pfd={-1,POLLIN,POLLIN};
    struct usb_event uev;
    SLIST_HEAD(,device_infos) attached_devices = SLIST_HEAD_INITIALIZER(attached_devices);
    pfd.fd = fd;

	dprint("-- Reading config file --\n");
	{
		struct usb_config_entry *e;
		SLIST_FOREACH( e, conf, next) {
			dprint("DESCRIPTION (%s)\n"
					"\tattach (%s)\n"
					"\tdetach (%s)\n"
					"\t * match items:\n",
					e->descr,e->attach,e->detach
				  );
			if( e->used_fields & USEFIELD_DEV )
				dprint("\t\tdev (%s)\n",e->dev);
			if( e->used_fields & USEFIELD_VENDOR_NO)
				dprint("\t\tvendorID (%X)\n",e->vendor_no);
			if( e->used_fields & USEFIELD_PRODUCT_NO)
				dprint("\t\tproductID (%X)\n",e->product_no);
			if( e->used_fields & USEFIELD_RELEASE_NO)
				dprint("\t\trelease (%X)\n",e->release_no);
			if( e->used_fields & USEFIELD_PRODUCT)
				dprint("\t\tproduct (%s)\n",e->product);
		}
	}
	dprint("-- EOF --\n");

    for(;;) {
        int ret;
        struct device_infos *dev_infos;

        if( (ret = poll(&pfd,1,INFTIM)) > 0) {
            read(fd,&uev,sizeof(struct usb_event));
            switch(uev.ue_type) {
                case USB_EVENT_DRIVER_ATTACH:
                    SLIST_FOREACH(dev_infos, &attached_devices, next) {
                        if( dev_infos->cookie.cookie == uev.u.ue_driver.ue_cookie.cookie ) {
                            strncpy(dev_infos->devname,uev.u.ue_driver.ue_devname,USB_MAX_DEVNAMELEN);
                            dprint("NEW DEVICE %s [%s]\n",dev_infos->product,dev_infos->devname);
							match_device(dev_infos,conf,ACTION_ATTACH);
                        }
                    }
                    break;
                case USB_EVENT_DRIVER_DETACH:
                    break;
                case USB_EVENT_DEVICE_ATTACH:
                    dev_infos = malloc(sizeof(struct device_infos));
                    memcpy( &(dev_infos->cookie) , &(uev.u.ue_device.udi_cookie.cookie) , sizeof(usb_event_cookie_t));
                    strncpy(dev_infos->product, uev.u.ue_device.udi_product, USB_MAX_STRING_LEN);
					dev_infos->product_no = uev.u.ue_device.udi_productNo;
					dev_infos->vendor_no = uev.u.ue_device.udi_vendorNo;
					dev_infos->release_no = uev.u.ue_device.udi_releaseNo;
                    SLIST_INSERT_HEAD( &attached_devices, dev_infos, next);
                    break;
                case USB_EVENT_DEVICE_DETACH:
                    SLIST_FOREACH(dev_infos, &attached_devices, next) {
                        if(uev.u.ue_device.udi_cookie.cookie == dev_infos->cookie.cookie ) {
                            dprint("DETACHED %s [%s]\n",dev_infos->product,dev_infos->devname);
							match_device(dev_infos,conf,ACTION_DETACH);
                            SLIST_REMOVE(&attached_devices,dev_infos,device_infos, next);
                            free(dev_infos);
                        }
                    }
                    break;
            }
        }
    }
}

void usage(char *progname)
{
	fprintf(stderr,
			"usage: %s [-d] [-f <filename>]\n"
			,progname);
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
    int fd = open(USB_DEVICE,O_RDONLY,0);
	struct usb_config_head conf;
	char *config_filename="pluged.conf";


	{
		int ch;
		extern char *optarg;
		extern int optind;

		while ((ch = getopt(argc, argv, "df:")) != -1) {
			switch (ch) {
				case 'd':
					debug++;
					break;
				case 'f':
					config_filename = optarg;
					break;
				case '?':
				default:
					usage(*argv);
			}
		}
		argc -= optind;

		/* optional parameter given */
		if(argc)
			usage(*argv);

		argv += optind;
	}

	conf = read_config(config_filename);

    if( -1 == fd ) {
        perror(USB_DEVICE);
        exit(EXIT_FAILURE);
    }

	if(debug == 0) {
		if(!fork()) {
			setsid();
			event_check_loop(fd, &conf);
			close(fd);
		}
	} else {
		event_check_loop(fd, &conf);
		close(fd);
	}

    return 0;
}
